from win10toast import ToastNotifier

def notify(title, subtitle, info_text, duration=5):
    ToastNotifier().show_toast(f"{title}: {subtitle}",
                               info_text,
                               icon_path=None,
                               duration=duration)



import io
import os
import sys

def spawn_daemon():
    # do the UNIX double-fork magic, see Stevens' "Advanced 
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try:
        r, w = os.pipe()
        pid = os.fork() 
        if pid > 0:
            # we are the parent process
            return int(os.fdopen(r).readline().strip())
    except OSError as e:
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(os.EX_OSERR)

    os.setsid()

    # do second fork
    try: 
        pid = os.fork()
        if pid > 0:
            # we are the parent process; let our parent know about our child's PID:
            os.write(w, f"{pid}\n".encode('utf8'))
            sys.exit(os.EX_OK)
    except OSError as e: 
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(os.EX_OSERR)
    return 0



from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import getpass
import os
import socket
import sys
import time

import keyring

from .notify import Notifier

KEYRING_NAME='NotifyWhenDone'

def prompt(message, yes_options = ('y', 'yes'), no_options = ('n', 'no'), default=True):
    if default:
        yes_options = yes_options + ('',)
    else:
        no_options = no_options + ('',)
    while True:
        ret = input(message).strip().lower()
        if ret in yes_options:
            return True
        elif ret in no_options:
            return False

def save_email_credentials(sender_name, sender_email, sender_password, smtp_server, smtp_port=587):
    keyring.set_password(KEYRING_NAME, 'sender_name', sender_name)
    keyring.set_password(KEYRING_NAME, 'sender_email', sender_email)
    keyring.set_password(KEYRING_NAME, 'sender_password', sender_password)
    keyring.set_password(KEYRING_NAME, 'smtp_server', smtp_server)
    keyring.set_password(KEYRING_NAME, 'smtp_port', str(smtp_port))

def prompt_for_email_credentials(sender_name=None, sender_email=None, sender_password=None, smtp_server=None, smtp_port=None):
    if sender_name is None:
        sender_name = keyring.get_password(KEYRING_NAME, 'sender_name')
    if sender_name is None:
        sender_name = getpass.getuser()
    new_sender_name = input(f"Sender Name: [{sender_name}] ")
    if new_sender_name == '':
        new_sender_name = sender_name

    if sender_email is None:
        sender_email = keyring.get_password(KEYRING_NAME, 'sender_email')
    while True:
        if sender_email is None:
            new_sender_email = input('Sender E-Mail: ')
        else:
            new_sender_email = input(f"Sender E-Mail: [{sender_email}] ")
            if new_sender_email == '':
                new_sender_email = sender_email
        if '@' in new_sender_email:
            break

    if sender_password is None:
        sender_password = keyring.get_password(KEYRING_NAME, 'sender_password')
    if sender_password is not None and new_sender_email == sender_email:
        if prompt(f"Would you like to update the password for {new_sender_email}? [yN] ", default=False):
            change_password = True
        else:
            change_password = False
            new_sender_password = sender_password
    else:
        change_password = True
    if change_password:
        new_sender_password = getpass.getpass(f"Password for {new_sender_email}: ")

    normalized_email = new_sender_email.strip().lower()
    if smtp_server is None:
        if normalized_email.endswith('@gmail.com'):
            smtp_server = 'smtp.gmail.com'
        elif normalized_email.endswith('@hotmail.com') or normalized_email.endswith('@live.com') or normalized_email.endswith('@outlook.com'):
            smtp_server = 'smtp-mail.outlook.com'
        else:
            smtp_server = keyring.get_password(KEYRING_NAME, 'smtp_server')
            if smtp_server is None:
                smtp_server = f"mail.{normalized_email[normalized_email.find('@')+1:]}"
    new_smtp_server = input(f"SMTP Server: [{smtp_server}] ")
    if new_smtp_server == '':
        new_smtp_server = smtp_server
    if new_smtp_server == 'smtp.gmail.com':
        sys.stderr.write('Note: If you have two factor authentication enabled on your Gmail account, you may need to set up an app password for NWD at https://security.google.com/settings/security/apppasswords\n')

    if smtp_port is None:
        smtp_port = keyring.get_password(KEYRING_NAME, 'smtp_port')
    if smtp_port is None:
        smtp_port = '587'
    new_smtp_port = input(f"SMTP Port: [{smtp_port}] ")
    if new_smtp_port == '':
        new_smtp_port = smtp_port

    return new_sender_name, new_sender_email, new_sender_password, new_smtp_server, int(new_smtp_port)

def prompt_and_save_email_credentials(*args, **kwargs):
    prev_creds = get_keychain_credentials()
    creds = prompt_for_email_credentials(*args, **kwargs)
    if creds[1] is None:
        raise Exception('Error loading e-mail credentials!')
    if prev_creds != creds and prompt('Save these e-mail credentials to the system keychain? [Yn] '):
        save_email_credentials(*creds)
    return creds

def get_keychain_credentials():
    sender_email = keyring.get_password(KEYRING_NAME, 'sender_email')
    sender_name = keyring.get_password(KEYRING_NAME, 'sender_name')
    sender_password = keyring.get_password(KEYRING_NAME, 'sender_password')
    smtp_server = keyring.get_password(KEYRING_NAME, 'smtp_server')
    smtp_port = keyring.get_password(KEYRING_NAME, 'smtp_port')
    if smtp_port is None:
        smtp_port = 587
    else:
        smtp_port = int(smtp_port)
    return sender_name, sender_email, sender_password, smtp_server, smtp_port    

def get_email_credentials(*args, **kwargs):
    creds = get_keychain_credentials()
    if creds[1] is None:
        return prompt_and_save_email_credentials(*args, **kwargs)
    else:
        return creds

def get_default_recipient():
    return keyring.get_password(KEYRING_NAME, 'default_recipient')

def save_default_recipient(recipient_email):
    keyring.set_password(KEYRING_NAME, 'default_recipient', recipient_email)

def get_recipient(default_recipient=None):
    default = get_default_recipient()
    if default is None:
        default = default_recipient
    while True:
        if default is None:
            recipient = input('E-Mail Address to Notify: ')
        else:
            recipient = input(f"E-Mail Address to Notify: [{default}] ")
            if recipient == '':
                recipient = default
        if '@' in recipient:
            break
    if recipient != default:
        if prompt(f"Would you like to make {recipient} the default notification recipient? [Yn] "):
            save_default_recipient(recipient)
    return recipient

class EmailNotifier(Notifier):
    def __init__(self, sender_name=None, sender_email=None, sender_password=None, smtp_server=None, smtp_port=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sender_name, self.sender_email, self.sender_password, self.smtp_server, self.smtp_port = get_email_credentials(sender_name, sender_email, sender_password, smtp_server, smtp_port)
        self.to_email = get_recipient(self.sender_email)

    def notify(self):
        # Need to import smtplib here, otherwise it will segfault if we import it before the fork()
        import smtplib
        msg = MIMEMultipart()
        msg['Subject'] = f"NWD: {self.name} finished!"
        msg['From'] = f"{self.sender_name} <{self.sender_email}>"
        msg['To'] = self.to_email
        if self.stdout or self.stderr:
            attachment_msg = '\n\nProgram output logs are attached.'
        else:
            attachment_msg = ''
        msg.attach(MIMEText(f"Process {self.pid} on {socket.gethostname()} finished at {time.ctime(self.end_time)}.\n\n`{' '.join(self.commandline)}`{attachment_msg}."))

        for logstream, name, mirror_to in ((self.stdout, 'stdout.txt'), (self.stderr, 'stderr.txt')):
            if logstream is not None:
                with open(logstream.name, "rb") as f:
                    part = MIMEApplication(
                        f.read(),
                        Name=name
                    )
                try:
                    os.unlink(logstream.name)
                except Exception as e:
                    print(e)
                part['Content-Disposition'] = f"attachment; filename=\"{name}\""
                msg.attach(part)

        server = smtplib.SMTP(f"{self.smtp_server}:{self.smtp_port}")
        server.starttls()
        server.login(self.sender_email, self.sender_password)
        server.sendmail(self.sender_email,
                        [self.to_email],
                        msg.as_string())
        server.quit()



def notify(title, subtitle, info_text, delay=0, sound=False, userInfo={}):
    # We need to import these here because they must be loaded from the child process after a fork
    import Foundation
    import objc
    import AppKit
    NSUserNotification = objc.lookUpClass('NSUserNotification')
    NSUserNotificationCenter = objc.lookUpClass('NSUserNotificationCenter')
    notification = NSUserNotification.alloc().init()
    notification.setTitle_(title)
    notification.setSubtitle_(subtitle)
    notification.setInformativeText_(info_text)
    notification.setUserInfo_(userInfo)
    if sound:
        notification.setSoundName_("NSUserNotificationDefaultSoundName")
    notification.setDeliveryDate_(Foundation.NSDate.dateWithTimeInterval_sinceDate_(delay, Foundation.NSDate.date()))
    NSUserNotificationCenter.defaultUserNotificationCenter().scheduleNotification_(notification)



from enum import Enum
import json
import os
from pathlib import Path
import sys
import time

import psutil

from .daemon import spawn_daemon

PID_DIR=os.path.join(Path.home(), '.nwd')
os.makedirs(PID_DIR, exist_ok=True)

class Status(Enum):
    NOT_STARTED = 0
    MONITORING  = 1
    NOTIFYING   = 2
    NOTIFIED    = 3
    FAILED      = 4

def get_notifiers():
    for pid in os.listdir(PID_DIR):
        path = os.path.join(PID_DIR, pid)
        if not os.path.isfile(path):
            continue
        try:
            pid = int(pid)
        except ValueError:
            continue
        with open(path) as f:
            proc_pid = json.load(f)['pid']
        yield Notifier(pid=proc_pid, daemon_pid=pid)

def cleanup():
    for notifier in get_notifiers():
        status = notifier.status
        if status == Status.NOTIFIED or status == Status.FAILED:
            os.unlink(notifier.pidfile)
            yield notifier, status

class Notifier(object):
    def __init__(self, pid, daemon_pid=None, stdout=None, stderr=None):
        self.pidfile = None
        self.pid = pid
        self._daemon_pid = None
        self.daemon_pid = daemon_pid
        self.stdout = stdout
        self.stderr = stderr

    @property
    def daemon_pid(self):
        return self._daemon_pid

    @daemon_pid.setter
    def daemon_pid(self, pid):
        if self._daemon_pid is not None:
            if self._daemon_pid == pid:
                return
            raise Exception(f"Cannot set the daemon PID to {pid} because it is already set to {self._daemon_pid}")
        self._daemon_pid = pid
        if pid is not None:
            self.pidfile = os.path.join(PID_DIR, str(pid))

    def _raw_status(self):
        if self.pidfile is None or not os.path.exists(self.pidfile):
            return {
                'pid' : self.pid,
                'name' : f"Process {self.pid}",
                'commandline' : [],
                'started' : None,
                'finished' : None,
                'status' : Status.NOT_STARTED.name
            }
        with open(self.pidfile) as f:
            return json.load(f)

    def _save_status(self, **kwargs):
        existing_status = self._raw_status()
        for k, v in kwargs.items():
            existing_status[k] = v
        with open(self.pidfile, 'w') as f:
            json.dump(existing_status, f)

    @property
    def status(self):
        s = Status.__members__[self._raw_status()['status']]
        try:
            if s != Status.NOTIFIED and not psutil.Process(self.daemon_pid).is_running():
                s = Status.FAILED
        except psutil.NoSuchProcess:
            if s != Status.NOTIFIED:
                s = Status.FAILED
        return s

    @status.setter
    def status(self, new_status):
        self._save_status(status=new_status.name)

    @property
    def name(self):
        return self._raw_status()['name']

    @property
    def commandline(self):
        return self._raw_status()['commandline']

    def notify(self):
        raise NotImplementedError('Subclasses of Notifier must implement the notify() function')

    @property
    def start_time(self):
        return self._raw_status()['started']

    @property
    def end_time(self):
        return self._raw_status()['finished']

    def terminate(self):
        status = self.status
        if status == Status.MONITORING or status == Status.NOTIFYING:
            psutil.wait_procs([psutil.Process(self.daemon_pid)])
        os.unlink(self.pidfile)

    def start(self, block=False):
        if self.daemon_pid is not None:
            return self.daemon_pid
        if not block:
            child_pid = spawn_daemon()
            if child_pid > 0:
                # we are in the parent process
                self.daemon_pid = child_pid
                return child_pid

        # We are now executing in the daemon process
        self.daemon_pid = os.getpid()
        self.status = Status.MONITORING
        # Maybe the process is already finished?
        process = psutil.Process(self.pid)
        if process.is_running():
            self._save_status(name=process.name(), commandline=process.cmdline(), started=process.create_time(), finished=time.time())
            process.wait()
        self.status = Status.NOTIFYING
        self.notify()
        self.status = Status.NOTIFIED
        sys.exit(os.EX_OK)



import argparse
from functools import partial
import os
import sys
import subprocess
import tempfile
import time

import psutil

from .daemon import spawn_daemon
from .desktop import DesktopNotifier
from .email import EmailNotifier, get_email_credentials, prompt_and_save_email_credentials
from .tabular import draw_table
from .term import TerminalNotifier
from . import notify

class RunNotifier(notify.Notifier):
    def __init__(self, run_command, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.run_command = run_command

    def notify(self):
        os.system(self.run_command)

def main(argv = None):
    parser = argparse.ArgumentParser(description='Notify When Done (NWD). A tool for posting a desktop notification, E-mail, or other alert when a process finishes.')
    parser.add_argument('PID', type=int, nargs='?', help='the process ID to monitor')
    parser.add_argument('--name', '-n', type=str, default=None, help='specify the process by its name instead of its PID')
    parser.add_argument('--exec', '-e', type=str, default=None, help='executes the given command creates a notification on its completion')
    parser.add_argument('--block', '-b', action='store_true', default=False, help='block until the monitored process terminates')
    parser.add_argument('--list', '-l', action='store_true', default=False, help='lists all pending and completed notifications')
    parser.add_argument('--delete', '-d', action='store_true', default=False, help='cancels the pending notification for the process specified by its PID or by --name')
    parser.add_argument('--cleanup', '-c', action='store_true', default=False, help='clean up stale output from --list')
    parser.add_argument('--mode', '-m', choices=['desktop', 'email', 'term'], default=None, help='sets the notification method, where \'desktop\' (the default) is a desktop notification popup, \'email\' sends an e-mail, and \'term\' prints a message to the terminal')
    parser.add_argument('--run', '-r', type=str, default=None, help='run the given command when the process completes')
    parser.add_argument('--email-credentials', action='store_true', default=False, help='prompt for the e-mail credentials with which to send alerts and offer to save them to the system keychain')

    if argv is None:
        argv = sys.argv
    
    args = parser.parse_args(argv[1:])

    num_run_args = sum(map(bool, (args.PID, args.name, args.exec)))
    if num_run_args > 1:
        parser.print_help(sys.stderr)
        sys.stderr.write('\nYou may only specify at most one of PID, --name, and --exec\n')
        sys.exit(1)
    elif num_run_args == 0 and not (args.cleanup or args.list or args.email_credentials):
        parser.print_help(sys.stderr)
        sys.exit(1)

    if args.delete:
        if not args.PID and args.name is None:
            parser.print_help(sys.stderr)
            sys.stderr.write('\nYou must specify either a PID or a --name when using --delete\n')
            sys.exit(1)
        elif args.mode is not None:
            parser.print_help(sys.stderr)
            sys.stderr.write('\nA notification --mode may not be specified when using --delete\n')
            sys.exit(1)
        elif args.exec is not None:
            parser.print_help(sys.stderr)
            sys.stderr.write('\nThe --exec option may not be specified when using --delete\n')
            sys.exit(1)
        elif args.run is not None:
            parser.print_help(sys.stderr)
            sys.stderr.write('\nThe --run option may not be specified when using --delete\n')
            sys.exit(1)

    if args.email_credentials:
        email_creds = prompt_and_save_email_credentials()
    elif args.mode == 'email':
        email_creds = get_email_credentials()

    if args.delete:
        for notifier in notify.get_notifiers():
            if (args.PID and notifier.pid == args.PID) or (args.name is not None and notifier.name == args.name):
                pid = notifier.pid
                daemonpid = notifier.daemon_pid
                notifier.terminate()
                print(f"Deleted NWD daemon PID {daemonpid} monitoring PID {pid}")

    if args.list:
        titles = ('PID', 'Name', 'Started', 'Ended', 'Status')
        data = tuple((str(n.pid), n.name, time.ctime(n.start_time), time.ctime(n.end_time), n.status.name) for n in notify.get_notifiers())
        if not data:
            sys.stderr.write('There are no processes monitored\n')
        else:
            draw_table(titles, data)

    if args.cleanup:
        for notifier, status in notify.cleanup():
            print(f"Cleaned up monitor {notifier.daemon_pid} for process {notifier.pid} ({status.name})")

    if args.delete:
        return
    elif args.mode is None and not args.exec and not args.PID and args.name is None and args.run is None:
        return

    stdout_w = None
    stderr_w = None
    if args.run is not None:
        Notifier = partial(RunNotifier, args.run)
    elif args.mode is None or args.mode == 'desktop':
        Notifier = DesktopNotifier
    elif args.mode == 'email':
        if args.exec:
            stdout_w = tempfile.NamedTemporaryFile(prefix='stdout', suffix='.txt', delete=False)
            stderr_w = tempfile.NamedTemporaryFile(prefix='stderr', suffix='.txt', delete=False)
        Notifier = partial(EmailNotifier, *email_creds, stdout=stdout_w, stderr=stderr_w)
    elif args.mode == 'term':
        Notifier = TerminalNotifier
    
    if args.exec:
        pid = spawn_daemon()
        if pid > 0:
            # parent process
            args.PID = pid
        else:
            # child process
            p = subprocess.Popen(args.exec.split(' '), stdout=stdout_w, stderr=stderr_w)
            sys.exit(p.wait())
    elif args.name is not None:
        matches = tuple(proc for proc in psutil.process_iter() if proc.name() == args.name and proc.pid != os.getpid())
        if not matches:
            sys.stderr.write(f"Error: Could not find a process named \"{args.name}\"\n")
            sys.exit(2)
        elif len(matches) > 1:
            sys.stderr.write(f"There are multiple processes running named \"{args.name}\":\n\n")
            titles = (' PID ', ' Command Line ')
            data = tuple((f" {match.pid} ", ' ' + ' '.join(match.cmdline()) + ' ') for match in matches)
            draw_table(titles, data)
            sys.stderr.write('\nPlease specify the one you want by its PID\n\n')
            sys.exit(3)
        args.PID = matches[0].pid

    try:
        monitor_pid = Notifier(args.PID).start(block=args.block)
    except Exception as e:
        sys.stderr.write(f"Error monitoring PID {args.PID}: {e}\n")
        if args.exec:
            sys.stderr.write('The execute command may have completed before we could attach an NWD monitor.\n')
        sys.exit(1)

    sys.stderr.write(f"Started monitoring daemon for PID {args.PID} at PID {monitor_pid}\n")
    
if __name__ == '__main__':
    main()



import sys

def draw_table(titles, data, out_stream=None, err_stream=None):
    if out_stream is None:
        out_stream = sys.stdout
    if err_stream is None:
        err_stream = sys.stderr
    
    colwidths = tuple(max(len(row[col]) for row in (titles,) + data) for col in range(len(titles)))
    rowwidth = sum(colwidths) + len(colwidths) - 1
    sys.stderr.write('|'.join(title.ljust(colwidths[col]) for col, title in enumerate(titles)))
    sys.stderr.write('\n')
    sys.stderr.write('+'.join('-' * width for width in colwidths))
    sys.stderr.write('\n')
    sys.stderr.flush()
    for row in data:
        for col, coldata in enumerate(row):
            assert len(coldata) <= colwidths[col]
            not_last = col < len(colwidths) - 1
            sys.stdout.write(coldata)
            if not_last:
                sys.stdout.write(', ')
            sys.stdout.flush()
            if not_last:
                sys.stderr.write('\b\b')
                sys.stderr.write(' ' * (colwidths[col] - len(coldata)))
                sys.stderr.write('|')
                sys.stderr.flush()
        sys.stdout.write('\n')
        sys.stdout.flush()



import os
from setuptools import setup, find_packages

setup(
    name='nwd',
    description='Notify When Done (NWD). A tool for posting a desktop notification, E-mail, or other alert when a process finishes.',
    url='https://github.com/ESultanik/nwd',
    author='Evan Sultanik',
    version='2.0.0',
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'pyobjc;platform_system=="Darwin"',
        'win10toast;platform_system=="Windows"',
        'keyring',
        'psutil'
    ],
    entry_points={
        'console_scripts': [
            'nwd = nwd.__main__:main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Topic :: System :: Monitoring',
        'Topic :: Utilities'
    ]
)



from enum import Enum
import sys
import time

from .notify import Notifier

if sys.platform.startswith('darwin'):
    from .osx import notify
elif sys.platform == 'win32' or sys.platform == 'cygwin':
    from .windows import notify
else:
    from .gtk import notify

class DesktopNotifier(Notifier):
    def notify(self):
        notify('nwd', f"{self.name} finished!", f"Process {self.pid} finished at {time.ctime(self.end_time)}")



try:
    import gi
except ModuleNotFoundError:
    raise ModuleNotFoundError("PyGObject is required to generate desktop notifications on your system. Please install it using the instructions here: https://pygobject.readthedocs.io/en/latest/getting_started.html")

gi.require_version('Gio', '2.0')

def notify(title, subtitle, info_text):
    from gi.repository import Gio

    NWD = Gio.Application.new("NWD", Gio.ApplicationFlags.FLAGS_NONE)
    NWD.register()

    notification = Gio.Notification.new(subtitle)
    notification.set_body(info_text)
    icon = Gio.ThemedIcon.new("dialog-information")
    notification.set_icon(icon)
    NWD.send_notification(None, notification)

